#include "baseSimulator.h"
#include "baseSimulator.hh"

#include "gtest/gtest.h"

#include "abstract/Architecture.test.cpp"
#include "abstract/Concentration.test.cpp"
#include "abstract/Continuous.test.cpp"
#include "abstract/Droplet.test.cpp"
#include "abstract/GradientGenerator.test.cpp"
#include "abstract/Membrane.test.cpp"
#include "abstract/Topology.test.cpp"

#include "hybrid/Concentration.test.cpp"
#include "hybrid/Continuous.test.cpp"
#include "hybrid/Opening.test.cpp"

#include "cfd/Concentration.test.cpp"
#include "cfd/Continuous.test.cpp"
